@@ -2,42 +2,69 @@
 Mount utilities.
 """
 
+from __future__ import absolute_import
+from __future__ import unicode_literals
+
 from collections import namedtuple
 from functools import partial
 import logging
-import sys
 
-from udiskie.common import wraps, setdefault
-from udiskie.compat import filter, basestring
-from udiskie.config import IgnoreDevice, FilterMatcher
-from udiskie.locale import _
+from .async_ import AsyncList, Coroutine, Return
+from .common import wraps, setdefault, exc_message
+from .compat import basestring
+from .config import IgnoreDevice, FilterMatcher
+from .locale import _
 
 
 __all__ = ['Mounter']
 
 
-def _device_method(fn):
+# TODO: add / remove / XXX_all should make proper use of the asynchronous
+# execution.
+
+
+@Coroutine.from_generator_function
+def _False():
+    yield Return(False)
+
+
+def _find_device(fn, set_error=False):
+    """
+    Decorator for Mounter methods taking a Device as their first argument.
+
+    Enables to pass the path name as first argument and does some common error
+    handling (logging).
+    """
     @wraps(fn)
     def wrapper(self, device_or_path, *args, **kwargs):
-        if isinstance(device_or_path, basestring):
-            device = self.udisks.find(device_or_path)
-            if device:
-                self._log.debug(_('found device owning "{0}": "{1}"',
-                                  device_or_path, device))
-            else:
-                self._log.error(_('no device found owning "{0}"',
-                                  device_or_path))
-                return False
-        else:
-            device = device_or_path
         try:
-            return fn(self, device, *args, **kwargs)
-        except device.Exception:
-            err = sys.exc_info()[1]
-            self._log.error(_('failed to {0} {1}: {2}',
-                              fn.__name__, device, err.message))
-            self._set_error(device, fn.__name__, err.message)
-            return False
+            device = self.udisks.find(device_or_path)
+        except ValueError as e:
+            self._log.error(exc_message(e))
+            return _False()
+        return Coroutine(fn(self, device, *args, **kwargs))
+    return wrapper
+
+
+def _sets_async_error(fn):
+    @wraps(fn)
+    def wrapper(self, device, *args, **kwargs):
+        async_ = fn(self, device, *args, **kwargs)
+        async_.errbacks.append(partial(self._error, fn, device))
+        return async_
+    return wrapper
+
+
+def _suppress_error(fn):
+    """
+    Prevent errors in this function from being shown. This is OK, since all
+    errors happen in sub-functions in which errors ARE logged.
+    """
+    @wraps(fn)
+    def wrapper(self, device, *args, **kwargs):
+        async_ = fn(self, device, *args, **kwargs)
+        async_.errbacks.append(lambda *args: True)
+        return async_
     return wrapper
 
 
@@ -68,7 +95,8 @@ class Mounter(object):
                  mount_options=None,
                  ignore_device=None,
                  prompt=None,
-                 browser=None):
+                 browser=None,
+                 cache=None):
         """
         Initialize mounter with the given defaults.
 
@@ -89,6 +117,7 @@ class Mounter(object):
             IgnoreDevice({'is_ignored': True, 'ignore': True})]
         self._prompt = prompt
         self._browser = browser
+        self._cache = cache
         self._log = logging.getLogger(__name__)
         try:
             # propagate error messages to UDisks1 daemon for 'Job failed'
@@ -97,7 +126,15 @@ class Mounter(object):
         except AttributeError:
             self._set_error = lambda device, action, message: None
 
-    @_device_method
+    def _error(self, fn, device, err, fmt):
+        message = exc_message(err)
+        self._log.error(_('failed to {0} {1}: {2}',
+                          fn.__name__, device, message))
+        self._set_error(device, fn.__name__, message)
+        return True
+
+    @_sets_async_error
+    @_find_device
     def browse(self, device):
         """
         Browse device.
@@ -108,17 +145,18 @@ class Mounter(object):
         """
         if not device.is_mounted:
             self._log.error(_("not browsing {0}: not mounted", device))
-            return False
+            yield Return(False)
         if not self._browser:
             self._log.error(_("not browsing {0}: no program", device))
-            return False
+            yield Return(False)
         self._log.debug(_('opening {0} on {0.mount_paths[0]}', device))
         self._browser(device.mount_paths[0])
         self._log.info(_('opened {0} on {0.mount_paths[0]}', device))
-        return True
+        yield Return(True)
 
     # mount/unmount
-    @_device_method
+    @_sets_async_error
+    @_find_device
     def mount(self, device):
         """
         Mount the device if not already mounted.
@@ -129,19 +167,20 @@ class Mounter(object):
         """
         if not self.is_handleable(device) or not device.is_filesystem:
             self._log.warn(_('not mounting {0}: unhandled device', device))
-            return False
+            yield Return(False)
         if device.is_mounted:
             self._log.info(_('not mounting {0}: already mounted', device))
-            return True
+            yield Return(True)
         fstype = str(device.id_type)
         options = self._mount_options(device)
         kwargs = dict(fstype=fstype, options=options)
         self._log.debug(_('mounting {0} with {1}', device, kwargs))
-        mount_path = device.mount(**kwargs)
+        mount_path = yield device.mount(**kwargs)
         self._log.info(_('mounted {0} on {1}', device, mount_path))
-        return True
+        yield Return(True)
 
-    @_device_method
+    @_sets_async_error
+    @_find_device
     def unmount(self, device):
         """
         Unmount a Device if mounted.
@@ -152,17 +191,18 @@ class Mounter(object):
         """
         if not self.is_handleable(device) or not device.is_filesystem:
             self._log.warn(_('not unmounting {0}: unhandled device', device))
-            return False
+            yield Return(False)
         if not device.is_mounted:
             self._log.info(_('not unmounting {0}: not mounted', device))
-            return True
+            yield Return(True)
         self._log.debug(_('unmounting {0}', device))
-        device.unmount()
+        yield device.unmount()
         self._log.info(_('unmounted {0}', device))
-        return True
+        yield Return(True)
 
     # unlock/lock (LUKS)
-    @_device_method
+    @_sets_async_error
+    @_find_device
     def unlock(self, device):
         """
         Unlock the device if not already unlocked.
@@ -173,23 +213,56 @@ class Mounter(object):
         """
         if not self.is_handleable(device) or not device.is_crypto:
             self._log.warn(_('not unlocking {0}: unhandled device', device))
-            return False
+            yield Return(False)
         if device.is_unlocked:
             self._log.info(_('not unlocking {0}: already unlocked', device))
-            return True
+            yield Return(True)
         if not self._prompt:
             self._log.error(_('not unlocking {0}: no password prompt', device))
-            return False
-        password = self._prompt(device)
+            yield Return(False)
+        unlocked = yield self._unlock_from_cache(device)
+        if unlocked:
+            yield Return(True)
+        password = yield self._prompt(device)
         if password is None:
             self._log.debug(_('not unlocking {0}: cancelled by user', device))
-            return False
+            yield Return(False)
         self._log.debug(_('unlocking {0}', device))
-        device.unlock(password)
+        yield device.unlock(password)
+        self._update_cache(device, password)
         self._log.info(_('unlocked {0}', device))
-        return True
+        yield Return(True)
 
-    @_device_method
+    @Coroutine.from_generator_function
+    def _unlock_from_cache(self, device):
+        if not self._cache:
+            yield Return(False)
+        try:
+            password = self._cache[device]
+        except KeyError:
+            yield Return(False)
+        self._log.debug(_('unlocking {0} using cached password', device))
+        try:
+            yield device.unlock(password)
+        except Exception:
+            self._log.debug(_('failed to unlock {0} using cached password', device))
+            yield Return(False)
+        self._log.debug(_('unlocked {0} using cached password', device))
+        yield Return(True)
+
+    def _update_cache(self, device, password):
+        if not self._cache:
+            return
+        self._cache[device] = password
+
+    def forget_password(self, device):
+        try:
+            del self._cache[device]
+        except KeyError:
+            pass
+
+    @_sets_async_error
+    @_find_device
     def lock(self, device):
         """
         Lock device if unlocked.
@@ -200,17 +273,18 @@ class Mounter(object):
         """
         if not self.is_handleable(device) or not device.is_crypto:
             self._log.warn(_('not locking {0}: unhandled device', device))
-            return False
+            yield Return(False)
         if not device.is_unlocked:
             self._log.info(_('not locking {0}: not unlocked', device))
-            return True
+            yield Return(True)
         self._log.debug(_('locking {0}', device))
-        device.lock()
+        yield device.lock()
         self._log.info(_('locked {0}', device))
-        return True
+        yield Return(True)
 
     # add/remove (unlock/lock or mount/unmount)
-    @_device_method
+    @_suppress_error
+    @_find_device
     def add(self, device, recursive=False):
         """
         Mount or unlock the device depending on its type.
@@ -221,24 +295,65 @@ class Mounter(object):
         :rtype: bool
         """
         if device.is_filesystem:
-            success = self.mount(device)
+            success = yield self.mount(device)
         elif device.is_crypto:
-            success = self.unlock(device)
+            success = yield self.unlock(device)
             if success and recursive:
                 # TODO: update device
-                success = self.add(device.luks_cleartext_holder,
-                                   recursive=True)
+                success = yield self.add(
+                    device.luks_cleartext_holder,
+                    recursive=True)
         elif recursive and device.is_partition_table:
-            success = True
+            tasks = []
             for dev in self.get_all_handleable():
                 if dev.is_partition and dev.partition_slave == device:
-                    success = self.add(dev, recursive=True) and success
+                    tasks.append(self.add(dev, recursive=True))
+            # TODO: AND results
+            success = yield AsyncList(tasks)
         else:
             self._log.info(_('not adding {0}: unhandled device', device))
-            return False
-        return success
+            yield Return(False)
+        yield Return(success)
 
-    @_device_method
+    @_suppress_error
+    @_find_device
+    def auto_add(self, device, recursive=False):
+        """
+        Automatically attempt to mount or unlock a device, but be quiet if the
+        device is not supported.
+
+        :param device: device object, block device path or mount path
+        :param bool recursive: recursively mount and unlock child devices
+        :returns: whether all attempted operations succeeded
+        :rtype: bool
+        """
+        success = True
+        if not self.is_handleable(device):
+            pass
+        elif device.is_filesystem:
+            if not device.is_mounted:
+                success = yield self.mount(device)
+        elif device.is_crypto:
+            if self._prompt and not device.is_unlocked:
+                success = yield self.unlock(device)
+            if success and recursive:
+                # TODO: update device
+                success = yield self.auto_add(
+                    device.luks_cleartext_holder,
+                    recursive=True)
+        elif recursive and device.is_partition_table:
+            tasks = []
+            for dev in self.get_all_handleable():
+                if dev.is_partition and dev.partition_slave == device:
+                    tasks.append(self.auto_add(dev, recursive=True))
+            # TODO: AND results
+            success = yield AsyncList(tasks)
+        else:
+            self._log.debug(_('not adding {0}: unhandled device', device))
+        yield Return(success)
+
+    @_suppress_error
+    @_find_device
     def remove(self, device, force=False, detach=False, eject=False,
                lock=False):
         """
@@ -253,35 +368,89 @@ class Mounter(object):
         :rtype: bool
         """
         if device.is_filesystem:
-            success = self.unmount(device)
+            success = yield self.unmount(device)
         elif device.is_crypto:
             if force and device.is_unlocked:
-                self.remove(device.luks_cleartext_holder, force=True)
-            success = self.lock(device)
+                yield self.auto_remove(device.luks_cleartext_holder, force=True)
+            success = yield self.lock(device)
         elif force and (device.is_partition_table or device.is_drive):
-            success = True
+            tasks = []
             for child in self.get_all_handleable():
                 if _is_parent_of(device, child):
-                    success = self.remove(child,
-                                          force=True,
-                                          detach=detach,
-                                          eject=eject,
-                                          lock=lock) and success
+                    tasks.append(self.auto_remove(
+                        child,
+                        force=True,
+                        detach=detach,
+                        eject=eject,
+                        lock=lock))
+            # TODO: AND results
+            success = yield AsyncList(tasks)
         else:
             self._log.info(_('not removing {0}: unhandled device', device))
             success = False
         # if these operations work, everything is fine, we can return True:
         if lock and device.is_luks_cleartext:
             device = device.luks_cleartext_slave
-            success = self.lock(device)
+            success = yield self.lock(device)
         if eject:
-            success = self.eject(device)
+            success = yield self.eject(device)
         if detach:
-            success = self.detach(device)
-        return success
+            success = yield self.detach(device)
+        yield Return(success)
+
+    @_suppress_error
+    @_find_device
+    def auto_remove(self, device, force=False, detach=False, eject=False,
+                    lock=False):
+        """
+        Unmount or lock the device depending on device type.
+
+        :param device: device object, block device path or mount path
+        :param bool force: recursively remove all child devices
+        :param bool detach: detach the root drive
+        :param bool eject: remove media from the root drive
+        :param bool lock: lock the associated LUKS cleartext slave
+        :returns: whether all attempted operations succeeded
+        :rtype: bool
+        """
+        success = True
+        if not self.is_handleable(device):
+            pass
+        elif device.is_filesystem:
+            if device.is_mounted:
+                success = yield self.unmount(device)
+        elif device.is_crypto:
+            if force and device.is_unlocked:
+                yield self.auto_remove(device.luks_cleartext_holder, force=True)
+            if device.is_unlocked:
+                success = yield self.lock(device)
+        elif force and (device.is_partition_table or device.is_drive):
+            tasks = []
+            for child in self.get_all_handleable():
+                if _is_parent_of(device, child):
+                    tasks.append(self.auto_remove(
+                        child,
+                        force=True,
+                        detach=detach,
+                        eject=eject,
+                        lock=lock))
+            # TODO: AND results
+            success = yield AsyncList(tasks)
+        else:
+            self._log.debug(_('not removing {0}: unhandled device', device))
+        # if these operations work, everything is fine, we can return True:
+        if lock and device.is_luks_cleartext:
+            device = device.luks_cleartext_slave
+            success = yield self.lock(device)
+        if eject and device.has_media:
+            success = yield self.eject(device)
+        if detach and device.is_detachable:
+            success = yield self.detach(device)
+        yield Return(success)
 
     # eject/detach device
-    @_device_method
+    @_sets_async_error
+    @_find_device
     def eject(self, device, force=False):
         """
         Eject a device after unmounting all its mounted filesystems.
@@ -293,19 +462,20 @@ class Mounter(object):
         """
         if not self.is_handleable(device):
             self._log.warn(_('not ejecting {0}: unhandled device'))
-            return False
+            yield Return(False)
         drive = device.drive
         if not (drive.is_drive and drive.is_ejectable):
             self._log.warn(_('not ejecting {0}: drive not ejectable', drive))
-            return False
+            yield Return(False)
         if force:
-            self.remove(drive, force=True)
+            yield self.auto_remove(drive, force=True)
         self._log.debug(_('ejecting {0}', device))
-        drive.eject()
+        yield drive.eject()
         self._log.info(_('ejected {0}', device))
-        return True
+        yield Return(True)
 
-    @_device_method
+    @_sets_async_error
+    @_find_device
     def detach(self, device, force=False):
         """
         Detach a device after unmounting all its mounted filesystems.
@@ -316,18 +486,18 @@ class Mounter(object):
         :rtype: bool
         """
         if not self.is_handleable(device):
-            self._log.warn(_('not detaching {0}: unhandled device'))
-            return False
+            self._log.warn(_('not detaching {0}: unhandled device', device))
+            yield Return(False)
         drive = device.root
         if not drive.is_detachable:
             self._log.warn(_('not detaching {0}: drive not detachable', drive))
-            return False
+            yield Return(False)
         if force:
-            self.remove(drive, force=True)
+            yield self.auto_remove(drive, force=True)
         self._log.debug(_('detaching {0}', device))
-        drive.detach()
+        yield drive.detach()
         self._log.info(_('detached {0}', device))
-        return True
+        yield Return(True)
 
     # mount_all/unmount_all
     def add_all(self, recursive=False):
@@ -338,13 +508,11 @@ class Mounter(object):
         :returns: whether all attempted operations succeeded
         :rtype: bool
         """
-        success = True
-        for device in self.get_all_handleable():
-            if (device.is_filesystem
-                    or device.is_crypto
-                    or (recursive and device.is_partition_table)):
-                success = self.add(device, recursive=recursive) and success
-        return success
+        tasks = []
+        for device in self.udisks:
+            tasks.append(self.auto_add(device, recursive=recursive))
+        # TODO: AND results
+        return AsyncList(tasks)
 
     def remove_all(self, detach=False, eject=False, lock=False):
         """
@@ -356,71 +524,18 @@ class Mounter(object):
         :returns: whether all attempted operations succeeded
         :rtype: bool
         """
-        success = True
+        tasks = []
         remove_args = dict(force=True, detach=detach, eject=eject, lock=lock)
         for device in self.get_all_handleable():
-            if (device.is_filesystem
-                    or device.is_crypto
-                    or device.is_partition_table
-                    or device.is_drive):
-                success = self.remove(device, **remove_args) and success
-        return success
-
-    def mount_all(self):
-        """
-        Mount handleable devices that are already present.
-
-        :returns: whether all attempted operations succeeded
-        :rtype: bool
-        """
-        success = True
-        for device in self.get_all_handleable():
-            success = self.mount(device) and success
-        return success
-
-    def unmount_all(self):
-        """
-        Unmount all filesystems handleable by udiskie.
-
-        :returns: whether all attempted operations succeeded
-        :rtype: bool
-        """
-        success = True
-        for device in self.get_all_handleable():
-            success = self.unmount(device) and success
-        return success
-
-    def eject_all(self, force=True):
-        """
-        Eject all ejectable devices.
-
-        :param bool force: remove child devices before trying to eject
-        :returns: whether all attempted operations succeeded
-        :rtype: bool
-        """
-        success = True
-        for device in self.get_all_handleable():
-            if device.is_drive and device.is_ejectable:
-                success = self.eject(device, force=force) and success
-        return success
-
-    def detach_all(self, force=True):
-        """
-        Detach all detachable devices.
-
-        :param bool force: remove child devices before trying to detach
-        :returns: whether all attempted operations succeeded
-        :rtype: bool
-        """
-        success = True
-        for device in self.get_all_handleable():
-            if device.is_drive and device.is_detachable:
-                success = self.detach(device, force=force) and success
-        return success
+            if device.parent_object_path != '/':
+                continue
+            tasks.append(self.auto_remove(device, **remove_args))
+        # TODO: AND results
+        return AsyncList(tasks)
 
     # iterate devices
-    @_device_method
     def is_handleable(self, device):
+        # TODO: handle pathes in first argument
         """
         Check whether this device should be handled by udiskie.
 
@@ -433,6 +548,38 @@ class Mounter(object):
         """
         return not self._ignore_device(device)
 
+    def is_addable(self, device):
+        """
+        Check if device can be added with ``auto_add``.
+        """
+        if not self.is_handleable(device):
+            return False
+        if device.is_filesystem:
+            return not device.is_mounted
+        if device.is_crypto:
+            return self._prompt and not device.is_unlocked
+        if device.is_partition_table:
+            return any(self.is_addable(dev)
+                       for dev in self.get_all_handleable()
+                       if dev.partition_slave == device)
+        return False
+
+    def is_removable(self, device):
+        """
+        Check if device can be removed with ``auto_remove``.
+        """
+        if not self.is_handleable(device):
+            return False
+        if device.is_filesystem:
+            return device.is_mounted
+        if device.is_crypto:
+            return device.is_unlocked
+        if device.is_partition_table or device.is_drive:
+            return any(self.is_removable(dev)
+                       for dev in self.get_all_handleable()
+                       if _is_parent_of(device, dev))
+        return False
+
     def get_all_handleable(self):
         """
         Enumerate all handleable devices currently known to udisks.
@@ -462,6 +609,7 @@ class DeviceActions(object):
         'lock': _('Lock {0}'),
         'eject': _('Eject {0}'),
         'detach': _('Unpower {0}'),
+        'forget_password': _('Clear password for {0}'),
     }
 
     def __init__(self, mounter, actions={}):
@@ -475,6 +623,7 @@ class DeviceActions(object):
             'lock': partial(mounter.remove, force=True),
             'eject': partial(mounter.eject, force=True),
             'detach': partial(mounter.detach, force=True),
+            'forget_password': mounter.forget_password,
         })
 
     def detect(self, root_device=''):
@@ -508,6 +657,9 @@ class DeviceActions(object):
                 yield 'lock'
             else:
                 yield 'unlock'
+            cache = self._mounter._cache
+            if cache and device in cache:
+                yield 'forget_password'
         if device.is_ejectable and device.has_media:
             yield 'eject'
         if device.is_detachable:
@@ -515,7 +667,7 @@ class DeviceActions(object):
 
     def _device_node(self, device):
         """Create an empty menu node for the specified device."""
-        label = device.id_label or device.device_presentation
+        label = device.ui_label
         # determine available methods
         methods = [Action(method, device,
                           self._labels[method].format(label),
