@@ -10,20 +10,30 @@ This wraps the DBus API of Udisks2 providing a common interface with the
 udisks1 module.
 """
 
+from __future__ import absolute_import
+from __future__ import unicode_literals
+
 from copy import copy, deepcopy
+from functools import partial
 import logging
 
 from gi.repository import GLib
 
-from udiskie.common import Emitter, samefile
-from udiskie.compat import filter
-from udiskie.dbus import DBusException, DBusService
-from udiskie.locale import _
+from .common import Emitter, samefile, AttrDictView, decode_ay
+from .compat import fix_str_conversions
+from .dbus import connect_service, MethodsProxy
+from .locale import _
+from .async_ import Coroutine, Return
 
-__all__ = ['Sniffer', 'Daemon']
+__all__ = ['Daemon']
 
 
 def object_kind(object_path):
+    """
+    Parse the kind of object from an UDisks2 object path.
+
+    Example: /org/freedesktop/UDisks2/block_devices/sdb1 => device
+    """
     try:
         return {
             'block_devices': 'device',
@@ -57,199 +67,74 @@ Interface = {
 }
 
 
-# ----------------------------------------
-# byte array to string conversion
-# ----------------------------------------
-
-try:
-    unicode
-except NameError:
-    unicode = str
-
-
-def decode(ay):
-    """Convert data from DBus queries to strings."""
-    if ay is None:
-        return ''
-    elif isinstance(ay, unicode):
-        return ay
-    elif isinstance(ay, bytes):
-        return ay.decode('utf-8')
-    else:
-        # dbus.Array([dbus.Byte]) or any similar sequence type:
-        return bytearray(ay).rstrip(bytearray((0,))).decode('utf-8')
-
-
-def encode(s):
-    """Convert data from DBus queries to strings."""
-    if s is None:
-        return b''
-    elif isinstance(s, unicode):
-        return s.encode('utf-8')
-    else:
-        return s
-
-
 # ----------------------------------------
 # Internal helper classes
 # ----------------------------------------
 
-class AttrDictView(object):
+class MethodHub(object):
 
-    """Provide attribute access view to a dictionary."""
+    """Provide MethodsProxies for queried interfaces of a DBus object."""
 
-    def __init__(self, data):
-        self.__data = data
+    def __init__(self, object_proxy):
+        """Initialize from (ObjectProxy)."""
+        self._object_proxy = object_proxy
 
     def __getattr__(self, key):
-        try:
-            return self.__data[key]
-        except KeyError:
-            raise AttributeError
+        """Return a MethodsProxy for the requested interface."""
+        return MethodsProxy(self._object_proxy, Interface[key])
 
 
-class OfflineProxy(object):
+class PropertyHub(object):
 
-    """
-    Provide offline attribute access to a single interface on a DBus object.
+    """Provide attribute accessors for queried interfaces of a DBus object."""
 
-    Object properties are accessed statically via table lookup.
-
-    This access method is to be preferred over the dynamic property lookup
-    in many cases as it is immune to a number of race conditions.
-    """
-
-    def __init__(self, proxy, interface, data):
-        """
-        Initialize wrapper.
-
-        :param DBusObject proxy: for dynamic property/method lookup
-        :param dict data: for static property lookup
-        """
-        self.property = AttrDictView(data)
-        self._proxy = proxy
-        self._interface = interface
-
-    @property
-    def method(self):
-        return self._proxy.get_interface(self._interface).method
-
-
-class OfflineInterfaceService(object):
-
-    """
-    Provide offline attribute access to multiple interfaces on a DBus object.
-
-    Method access is performed dynamically via the given DBus proxy object.
-    """
-
-    def __init__(self, proxy, data):
-        """
-        Store DBus proxy and static property values.
-
-        :param dbus.proxies.ProxyObject proxy: DBus object for method access
-        :param dict data: interface and their properties a{sa{sv}}
-        """
-        self._proxy = proxy
-        self._data = data
+    def __init__(self, interfaces_and_properties):
+        """Initialize from (dict)."""
+        self._interfaces_and_properties = interfaces_and_properties
 
     def __getattr__(self, key):
-        """Return a wrapper for the requested interface."""
+        """Return an AttrDictView for properties on the requested interface."""
         interface = Interface[key]
         try:
-            return OfflineProxy(self._proxy, interface,
-                                self._data[interface])
-        except:
-            return NullProxy(key, self._proxy.object_path)
+            return AttrDictView(self._interfaces_and_properties[interface])
+        except KeyError:
+            return PropertiesNotAvailable()
 
 
-class OnlineInterfaceService(object):
+class PropertiesNotAvailable(object):
 
-    """
-    Provide online attribute access to multiple interfaces on a DBus object.
-
-    Both method and property access is performed dynamically via the given
-    DBus proxy object.
-    """
-
-    def __init__(self, proxy):
-        """
-        Store DBus proxy.
-
-        :param DBusObject object: DBus object for online access
-        """
-        self._proxy = proxy
-        properties = proxy.get_interface(Interface['Properties'])
-        self._check = properties.method.GetAll
-
-    def __getattr__(self, key):
-        """Return a wrapper for the requested interface."""
-        try:
-            self._check('(s)', Interface[key])
-            return self._proxy.get_interface(Interface[key])
-        except DBusException:
-            return NullProxy(key, self._proxy.object_path)
-
-    # TODO: need reliable and fast __nonzero__ check
-
-
-class NoneServer(object):
-
-    """Yield None when asked for any attribute."""
-
-    def __getattr__(self, key):
-        return None
-
-
-class NullProxy(object):
-
-    """Interface not available."""
-
-    def __init__(self, name, object_path):
-        self.object_path = object_path
-        self._name = name
-        self.property = NoneServer()
+    """Null class for properties of an unavailable interface."""
 
     def __nonzero__(self):      # python2
         return False
     __bool__ = __nonzero__      # python3
 
-    @property
-    def method(self):
-        """Access object methods dynamically via DBus."""
-        raise RuntimeError(_("Interface {0!r} not available for {1}",
-                             self._name, self.object_path))
+    def __getattr__(self, key):
+        """Return None when asked for any attribute."""
+        return None
 
 
 # ----------------------------------------
 # Device wrapper
 # ----------------------------------------
 
+@fix_str_conversions
 class Device(object):
 
     """
-    Wrapper class for DBus API objects representing devices.
+    Proxy class for UDisks2 devices.
 
-    Properties can be resolved dynamically through DBus or via table lookup
-    by providing an appropriate interface_service in the constructor. See
-    ``OfflineInterfaceService``.
-
-    This class is intended to be used only internally.
+    Properties are read from the cached values retrieved by the Daemon class.
+    Methods are executed asynchronously, and hence return Asyncs instead of
+    returning the result directly.
     """
 
-    Exception = DBusException
-
-    def __init__(self, udisks, object_path, interface_service):
-        """
-        Initialize an instance with the given DBus proxy object.
-
-        :param UDisks2 udisks: used to create other Device instances
-        :param str object_path: object path of the device
-        :param InterfaceService interface_service: used to access DBus API
-        """
-        self._udisks = udisks
+    def __init__(self, daemon, object_path, property_hub, method_hub):
+        """Initialize from (Daemon, str, PropertyHub, MethodHub)."""
+        self._daemon = daemon
         self.object_path = object_path
-        self._I = interface_service
+        self._P = property_hub
+        self._M = method_hub
 
     def __str__(self):
         """Show as object_path."""
@@ -263,35 +148,26 @@ class Device(object):
         """Comparison by object_path."""
         return not (self == other)
 
-    def __nonzero__(self):      # python2
-        return self.is_valid
-    __bool__ = __nonzero__      # python3
-
     def is_file(self, path):
         """Comparison by mount and device file path."""
         return samefile(path, self.device_file) or any(
             samefile(path, mp) for mp in self.mount_paths)
 
     # availability of interfaces
-    @property
-    def is_valid(self):
-        """Check if any interface is available for this object path."""
-        return bool(self._I)
-
     @property
     def is_drive(self):
         """Check if the device is a drive."""
-        return bool(self._I.Drive)
+        return bool(self._P.Drive)
 
     @property
     def is_block(self):
         """Check if the device is a block device."""
-        return bool(self._I.Block)
+        return bool(self._P.Block)
 
     @property
     def is_partition_table(self):
         """Check if the device is a partition table."""
-        return bool(self._I.PartitionTable)
+        return bool(self._P.PartitionTable)
 
     @property
     def is_partition(self):
@@ -299,17 +175,17 @@ class Device(object):
         # Sometimes udisks2 empties the Partition interface before removing
         # the device. In this case, we want to report .is_partition=False, so
         # properties like .partition_slave will not be used.
-        return bool(self._I.Partition and self.partition_slave)
+        return bool(self._P.Partition and self.partition_slave)
 
     @property
     def is_filesystem(self):
         """Check if the device is a filesystem."""
-        return bool(self._I.Filesystem)
+        return bool(self._P.Filesystem)
 
     @property
     def is_luks(self):
         """Check if the device is a LUKS container."""
-        return bool(self._I.Encrypted)
+        return bool(self._P.Encrypted)
 
     # ----------------------------------------
     # Drive
@@ -334,22 +210,22 @@ class Device(object):
     @property
     def is_detachable(self):
         """Check if the drive that owns this device can be detached."""
-        return bool(self._assocdrive._I.Drive.property.CanPowerOff)
+        return bool(self._assocdrive._P.Drive.CanPowerOff)
 
     @property
     def is_ejectable(self):
         """Check if the drive that owns this device can be ejected."""
-        return bool(self._assocdrive._I.Drive.property.Ejectable)
+        return bool(self._assocdrive._P.Drive.Ejectable)
 
     @property
     def has_media(self):
         """Check if there is media available in the drive."""
-        return bool(self._assocdrive._I.Drive.property.MediaAvailable)
+        return bool(self._assocdrive._P.Drive.MediaAvailable)
 
     # Drive methods
     def eject(self, auth_no_user_interaction=None):
         """Eject media from the device."""
-        return self._assocdrive._I.Drive.method.Eject(
+        return self._assocdrive._M.Drive.Eject(
             '(a{sv})',
             filter_opt({
                 'auth.no_user_interaction': ('b', auth_no_user_interaction),
@@ -358,7 +234,7 @@ class Device(object):
 
     def detach(self, auth_no_user_interaction=None):
         """Detach the device by e.g. powering down the physical port."""
-        return self._assocdrive._I.Drive.method.PowerOff(
+        return self._assocdrive._M.Drive.PowerOff(
             '(a{sv})',
             filter_opt({
                 'auth.no_user_interaction': ('b', auth_no_user_interaction),
@@ -373,22 +249,22 @@ class Device(object):
     @property
     def device_file(self):
         """The filesystem path of the device block file."""
-        return decode(self._I.Block.property.Device)
+        return decode_ay(self._P.Block.Device)
 
     @property
     def device_presentation(self):
         """The device file path to present to the user."""
-        return decode(self._I.Block.property.PreferredDevice)
+        return decode_ay(self._P.Block.PreferredDevice)
 
     @property
     def device_size(self):
         """The size of the device in bytes."""
-        return self._I.Block.property.Size
+        return self._P.Block.Size
 
     @property
     def id_usage(self):
         """Device usage class, for example 'filesystem' or 'crypto'."""
-        return decode(self._I.Block.property.IdUsage)
+        return self._P.Block.IdUsage
 
     @property
     def is_crypto(self):
@@ -398,7 +274,7 @@ class Device(object):
     @property
     def is_ignored(self):
         """Check if the device should be ignored."""
-        return self._I.Block.property.HintIgnore
+        return self._P.Block.HintIgnore
 
     @property
     def device_id(self):
@@ -409,12 +285,12 @@ class Device(object):
         `/dev/disk/by-id/`.
         """
         if self.is_block:
-            for filename in self._I.Block.property.Symlinks:
-                parts = decode(filename).split('/')
+            for filename in self._P.Block.Symlinks:
+                parts = decode_ay(filename).split('/')
                 if parts[-2] == 'by-id':
                     return parts[-1]
         elif self.is_drive:
-            return decode(self._assocdrive._I.Drive.property.Id)
+            return self._assocdrive._P.Drive.Id
         return ''
 
     @property
@@ -427,22 +303,22 @@ class Device(object):
         IdUsage     'filesystem'    'crypto'
         IdType      'ext4'          'crypto_LUKS'
         """
-        return decode(self._I.Block.property.IdType)
+        return self._P.Block.IdType
 
     @property
     def id_label(self):
         """Label of the device if available."""
-        return decode(self._I.Block.property.IdLabel)
+        return self._P.Block.IdLabel
 
     @property
     def id_uuid(self):
         """Device UUID."""
-        return decode(self._I.Block.property.IdUUID)
+        return self._P.Block.IdUUID
 
     @property
     def luks_cleartext_slave(self):
         """Get wrapper to the LUKS crypto device."""
-        return self._udisks[self._I.Block.property.CryptoBackingDevice]
+        return self._daemon[self._P.Block.CryptoBackingDevice]
 
     @property
     def is_luks_cleartext(self):
@@ -454,7 +330,7 @@ class Device(object):
         """Check if the device is external."""
         # NOTE: Checking for equality HintSystem==False returns False if the
         # property is resolved to a None value (interface not available).
-        if self._I.Block.property.HintSystem == False:
+        if self._P.Block.HintSystem == False:
             return True
         # NOTE: udisks2 seems to guess incorrectly in some cases. This
         # leads to HintSystem=True for unlocked devices. In order to show
@@ -480,7 +356,7 @@ class Device(object):
         if cleartext:
             return cleartext.drive
         if self.is_block:
-            return self._udisks[self._I.Block.property.Drive]
+            return self._daemon[self._P.Block.Drive]
         return None
 
     @property
@@ -489,7 +365,7 @@ class Device(object):
         Get the top level block device in the ancestry of this device.
         """
         drive = self.drive
-        for device in self._udisks:
+        for device in self._daemon:
             if device.is_drive:
                 continue
             if device.is_toplevel and device.drive == drive:
@@ -499,17 +375,17 @@ class Device(object):
     @property
     def should_automount(self):
         """Check if the device should be automounted."""
-        return bool(self._I.Block.property.HintAuto)
+        return bool(self._P.Block.HintAuto)
 
     @property
     def icon_name(self):
         """Return the recommended device icon name."""
-        return self._I.Block.property.HintIconName or 'drive-removable-media'
+        return self._P.Block.HintIconName or 'drive-removable-media'
 
     @property
     def symbolic_icon_name(self):
         """Return the recommended device symbolic icon name."""
-        return self._I.Block.property.HintSymbolicIconName or 'drive-removable-media'
+        return self._P.Block.HintSymbolicIconName or 'drive-removable-media'
 
     # ----------------------------------------
     # Partition
@@ -519,7 +395,7 @@ class Device(object):
     @property
     def partition_slave(self):
         """Get the partition slave (container)."""
-        return self._udisks[self._I.Partition.property.Table]
+        return self._daemon[self._P.Partition.Table]
 
     # ----------------------------------------
     # Filesystem
@@ -529,12 +405,12 @@ class Device(object):
     @property
     def is_mounted(self):
         """Check if the device is mounted."""
-        return bool(self._I.Filesystem.property.MountPoints)
+        return bool(self._P.Filesystem.MountPoints)
 
     @property
     def mount_paths(self):
         """Return list of active mount paths."""
-        return list(map(decode, self._I.Filesystem.property.MountPoints or ()))
+        return list(map(decode_ay, self._P.Filesystem.MountPoints or ()))
 
     # Filesystem methods
     def mount(self,
@@ -542,7 +418,7 @@ class Device(object):
               options=None,
               auth_no_user_interaction=None):
         """Mount filesystem."""
-        return self._I.Filesystem.method.Mount(
+        return self._M.Filesystem.Mount(
             '(a{sv})',
             filter_opt({
                 'fstype': ('s', fstype),
@@ -553,7 +429,7 @@ class Device(object):
 
     def unmount(self, force=None, auth_no_user_interaction=None):
         """Unmount filesystem."""
-        return self._I.Filesystem.method.Unmount(
+        return self._M.Filesystem.Unmount(
             '(a{sv})',
             filter_opt({
                 'force': ('b', force),
@@ -571,7 +447,7 @@ class Device(object):
         """Get wrapper to the unlocked luks cleartext device."""
         if not self.is_luks:
             return None
-        for device in self._udisks:
+        for device in self._daemon:
             if device.luks_cleartext_slave == self:
                 return device
         return None
@@ -584,21 +460,17 @@ class Device(object):
     # Encrypted methods
     def unlock(self, password, auth_no_user_interaction=None):
         """Unlock Luks device."""
-        object_path = self._I.Encrypted.method.Unlock(
+        return self._M.Encrypted.Unlock(
             '(sa{sv})',
             password,
             filter_opt({
                 'auth.no_user_interaction': ('b', auth_no_user_interaction),
             })
         )
-        # UDisks2 may not have processed the InterfacesAdded signal yet.
-        # Therefore it is necessary to query the interface data directly
-        # from the DBus service:
-        return self._udisks.update(object_path)
 
     def lock(self, auth_no_user_interaction=None):
         """Lock Luks device."""
-        return self._I.Encrypted.method.Lock(
+        return self._M.Encrypted.Lock(
             '(a{sv})',
             filter_opt({
                 'auth.no_user_interaction': ('b', auth_no_user_interaction),
@@ -615,91 +487,27 @@ class Device(object):
         if self.is_mounted or self.is_unlocked:
             return True
         if self.is_partition_table:
-            for device in self._udisks:
+            for device in self._daemon:
                 if device.partition_slave == self and device.in_use:
                     return True
         return False
 
+    @property
+    def parent_object_path(self):
+        return (self._P.Partition.Table
+                or self._P.Block.CryptoBackingDevice
+                or '/')
+
+    @property
+    def ui_label(self):
+        return self.id_label or self.id_uuid or self.device_presentation
+
 
 # ----------------------------------------
 # UDisks2 service wrapper
 # ----------------------------------------
 
-class UDisks2(DBusService):
-
-    """
-    Base class for UDisks2 service wrappers.
-    """
-
-    BusName = 'org.freedesktop.UDisks2'
-    ObjectPath = '/org/freedesktop/UDisks2'
-    Interface = Interface['ObjectManager']
-
-    def __iter__(self):
-        """Iterate over all devices."""
-        return filter(None, (self[path] for path in self.paths()
-                             if object_kind(path) in ('device', 'drive')))
-
-    def __getitem__(self, object_path):
-        return self.get(object_path)
-
-    def find(self, path):
-        """
-        Get a device proxy by device name or any mount path of the device.
-
-        This searches through all accessible devices and compares device
-        path as well as mount pathes.
-        """
-        for device in self:
-            if device.is_file(path):
-                return device
-        logger = logging.getLogger(__name__)
-        logger.warn(_('Device not found: {0}', path))
-        return None
-
-
-class Sniffer(UDisks2):
-    """
-    UDisks2 DBus service wrapper.
-
-    This is a wrapper for the DBus API of the UDisks2 service at
-    'org.freedesktop.UDisks2'. Access to properties and device states is
-    completely online, meaning the properties are requested from dbus as
-    they are accessed in the python object.
-    """
-
-    # Construction
-    def __init__(self, proxy=None):
-        """
-        Initialize an instance with the given DBus proxy object.
-
-        :param dbus.Bus bus: connection to system bus
-        :param common.DBusProxy proxy: proxy to udisks object
-        """
-        self._proxy = proxy or self.connect_service()
-        # Make sure the proxy object is loaded and usable:
-        self._proxy.method.GetManagedObjects()
-
-    # instantiation of device objects
-    def paths(self):
-        return self._proxy.method.GetManagedObjects().keys()
-
-    def _is_valid_object_path(self, object_path):
-        return (object_path
-                and object_path.startswith(self.ObjectPath)
-                and object_kind(object_path))
-
-    def get(self, object_path):
-        """Create a Device instance from object path."""
-        if not self._is_valid_object_path(object_path):
-            return None
-        return Device(self, object_path, OnlineInterfaceService(
-            self._proxy.object.bus.get_object(object_path)))
-
-    update = get
-
-
-class Daemon(Emitter, UDisks2):
+class Daemon(Emitter):
 
     """
     Listen to state changes to provide automatic synchronization.
@@ -714,7 +522,35 @@ class Daemon(Emitter, UDisks2):
         - device_changed  / job_failed
     """
 
-    def __init__(self, proxy=None):
+    BusName = 'org.freedesktop.UDisks2'
+    ObjectPath = '/org/freedesktop/UDisks2'
+    Interface = Interface['ObjectManager']
+
+    def __iter__(self):
+        """Iterate over all devices."""
+        return (self[path] for path in self.paths()
+                if object_kind(path) in ('device', 'drive'))
+
+    def __getitem__(self, object_path):
+        return self.get(object_path)
+
+    def find(self, path):
+        """
+        Get a device proxy by device name or any mount path of the device.
+
+        This searches through all accessible devices and compares device
+        path as well as mount pathes.
+        """
+        if isinstance(path, Device):
+            return path
+        for device in self:
+            if device.is_file(path):
+                self._log.debug(_('found device owning "{0}": "{1}"',
+                                  path, device))
+                return device
+        raise ValueError(_('no device found owning "{0}"', path))
+
+    def __init__(self, proxy):
 
         """Initialize object and start listening to UDisks2 events."""
 
@@ -730,7 +566,7 @@ class Daemon(Emitter, UDisks2):
                        'job_failed']
         super(Daemon, self).__init__(event_names)
 
-        self._proxy = proxy = proxy or self.connect_service()
+        self._proxy = proxy
         self._log = logging.getLogger(__name__)
         self._objects = {}
 
@@ -746,11 +582,22 @@ class Daemon(Emitter, UDisks2):
                     'Completed',
                     None,
                     self._job_completed)
-        self._sync()
 
     def _sync(self):
         """Synchronize state."""
-        self._objects = self._proxy.method.GetManagedObjects()
+        def update_objects(objects):
+            self._objects = objects
+        update = self._proxy.call('GetManagedObjects', '()')
+        update.callbacks.append(update_objects)
+        return update
+
+    @classmethod
+    @Coroutine.from_generator_function
+    def create(cls):
+        proxy = yield connect_service(cls)
+        daemon = cls(proxy)
+        yield daemon._sync()
+        yield Return(daemon)
 
     # UDisks2 interface
     def paths(self):
@@ -764,14 +611,10 @@ class Daemon(Emitter, UDisks2):
             interfaces_and_properties = self._objects.get(object_path)
             if not interfaces_and_properties:
                 return None
-        interface_service = OfflineInterfaceService(
-            self._proxy.object.bus.get_object(object_path),
-            interfaces_and_properties)
-        return Device(self, object_path, interface_service)
-
-    def update(self, object_path):
-        return self.get(object_path,
-                        self._proxy.method.GetManagedObjects()[object_path])
+        property_hub = PropertyHub(interfaces_and_properties)
+        method_hub = MethodHub(
+            self._proxy.object.bus.get_object(object_path))
+        return Device(self, object_path, property_hub, method_hub)
 
     def trigger(self, event, device, *args):
         self._log.debug(_("+++ {0}: {1}", event, device))
@@ -781,7 +624,10 @@ class Daemon(Emitter, UDisks2):
     def _interfaces_added(self, object_path, interfaces_and_properties):
         """Internal method."""
         added = object_path not in self._objects
-        self._objects[object_path] = interfaces_and_properties
+        self._objects.setdefault(object_path, {})
+        old_state = copy(self._objects[object_path])
+        self._objects[object_path].update(interfaces_and_properties)
+        new_state = self._objects[object_path]
         if added:
             kind = object_kind(object_path)
             if kind in ('device', 'drive'):
@@ -793,6 +639,11 @@ class Daemon(Emitter, UDisks2):
                 if not self._has_job(slave.object_path, 'device_unlocked'):
                     self.trigger('device_unlocked', slave)
 
+        if not added:
+            self.trigger('device_changed',
+                         self.get(object_path, old_state),
+                         self.get(object_path, new_state))
+
     # remove objects / interfaces
     def _detect_toggle(self, property_name, old, new, add_name, del_name):
         old_valid = old and bool(getattr(old, property_name))
@@ -811,8 +662,8 @@ class Daemon(Emitter, UDisks2):
                 job = interfaces[job_interface]
                 job_objects = job['Objects']
                 job_operation = job['Operation']
-                job_action = self._action_mapping[job_operation]
-                job_event = self._event_mapping[job_action]
+                job_action = self._action_by_operation[job_operation]
+                job_event = self._event_by_action[job_action]
                 if event_name == job_event and device_path in job_objects:
                     return True
             except KeyError:
@@ -839,7 +690,11 @@ class Daemon(Emitter, UDisks2):
                 if not self._has_job(slave.object_path, 'device_locked'):
                     self.trigger('device_locked', slave)
 
-        if not self._objects[object_path]:
+        if self._objects[object_path]:
+            self.trigger('device_changed',
+                         self.get(object_path, old_state),
+                         self.get(object_path, new_state))
+        else:
             del self._objects[object_path]
             if object_kind(object_path) in ('device', 'drive'):
                 self.trigger(
@@ -877,12 +732,15 @@ class Daemon(Emitter, UDisks2):
                 self.get(object_path, old_state),
                 self.get(object_path, new_state),
                 'device_mounted', 'device_unmounted')
+        self.trigger('device_changed',
+                     self.get(object_path, old_state),
+                     self.get(object_path, new_state))
         # There is no PropertiesChanged for the crypto device when it is
         # unlocked/locked in UDisks2. Instead, this is handled by the
         # InterfaceAdded/Removed handlers.
 
     # jobs
-    _action_mapping = {
+    _action_by_operation = {
         'filesystem-mount': 'mount',
         'filesystem-unmount': 'unmount',
         'encrypted-unlock': 'unlock',
@@ -891,7 +749,7 @@ class Daemon(Emitter, UDisks2):
         'eject-media': 'eject',
     }
 
-    _event_mapping = {
+    _event_by_action = {
         'mount': 'device_mounted',
         'unmount': 'device_unmounted',
         'unlock': 'device_unlocked',
@@ -900,7 +758,7 @@ class Daemon(Emitter, UDisks2):
         'detach': 'device_removed',
     }
 
-    _check_success = {
+    _check_action_success = {
         'mount': lambda dev: dev.is_mounted,
         'unmount': lambda dev: not dev or not dev.is_mounted,
         'unlock': lambda dev: dev.is_unlocked,
@@ -916,7 +774,7 @@ class Daemon(Emitter, UDisks2):
         Called when a job of a long running task completes.
         """
         job = self._objects[job_name][Interface['Job']]
-        action = self._action_mapping.get(job['Operation'])
+        action = self._action_by_operation.get(job['Operation'])
         if not action:
             return
         # We only handle events, which are associated to exactly one object:
@@ -926,8 +784,8 @@ class Daemon(Emitter, UDisks2):
             # It rarely happens, but sometimes UDisks posts the
             # Job.Completed event before PropertiesChanged, so we have to
             # check if the operation has been carried out yet:
-            if self._check_success[action](device):
-                event_name = self._event_mapping[action]
+            if self._check_action_success[action](device):
+                event_name = self._event_by_action[action]
                 self.trigger(event_name, device)
         else:
             self.trigger('job_failed', device, action, message)
